#!/usr/bin/env python3

__author__ = "Ahmed Hermas"
__copyright__ = "Copyright 2021, © Digified "
__license__ = "MIT"
__version__ = "0.1.0"
__email__ = "a7medhermas@gmail.com"

from face_utils.detection.detectors import Detector



#!/usr/bin/env python3

"""settings.py: detection module settings file """

__author__ = "Ahmed Hermas"
__copyright__ = "Copyright 2021, © Digified "
__license__ = "MIT"
__version__ = "0.1.0"
__email__ = "a7medhermas@gmail.com"

MODELS_CONFIG = {
    'hog' : 
        {'loader':'_hog_cnn_loader',
         'detector':'_getface_hog_cnn'},
    'cnn':
        {'loader':'_hog_cnn_loader',
         'detector':'_getface_hog_cnn'},
    'retina':
        {'loader':'_retina_loader',
         'detector':'_getface_retina'},
    'cascade':
        {'loader':'_cascade_loader',
         'detector':'_getface_cascade'},
    
}


import argparse
import os
import numpy as np
from cv2 import cv2
from face_utils.detection import Detector
from face_utils.cropping import cropping


ap = argparse.ArgumentParser()
ap.add_argument("-i", "--image", required = False, help = "Path to the input image")
ap.add_argument("-d", "--model", required = False, help = "detector model",default='hog')
ap.add_argument("-s", "--size", required = False, help = "the output size",default=128,type=int)
ap.add_argument("-o", "--out", required = True, help = "Path to out folder")
ap.add_argument("-f", "--folder", required = False, help = "Path to input folder")


args = ap.parse_args()

def detectImage(inPath,outFolder,model,size):
    
    image = cv2.imread(inPath)
    img= np.array(image)
    bbox = model.detect(img,2) #using (x,y,w,h) return mode
    outImg = cropping.crop(img,bbox,1,size,size)
    outFile = os.path.basename(inPath).split('.')[0]+"_out.jpg"
    outPath = os.path.join(outFolder,outFile)
    cv2.imwrite(outPath,outImg)
    
if __name__ == "__main__":
    
    modelName = args.model
    model = Detector(modelName)
    outFolder = args.out
    size = args.size
    
    if 'image' in args : 
        imgPath = args.image
        detectImage(imgPath,outFolder,model,size)
        
    else: 
        for file in os.listdir(args.folder):
            imgPath = os.path.join(args.folder,file)
            detectImage(imgPath,outFolder,model,size)


from setuptools import setup, find_packages

with open("README.md", mode="r", encoding="utf-8") as readme_file:
    readme = readme_file.read()


setup(
    name="face-utils",
    version="0.1.0",
    include_package_data=True,
    author="Hermas",
    author_email="a7medhermas@gmail.com",
    url="https://github.com/HermasTV/face_utils",
    description="opensource wrapper library for face detection algorithms and other utilities",
    long_description=readme,
    long_description_content_type="text/markdown",
    license="MIT",
    packages=find_packages(),
    install_requires=[
        'cmake',
        'opencv-python',
        'face-recognition'
    ],
    classifiers=[
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python :: 3.7",
        "Topic :: Scientific/Engineering :: Artificial Intelligence "
    ],
    keywords="face_utils face-utils hermas digified python face utils datection"
)



#!/usr/bin/env python3

"""cropping.py: the image cropping module."""

__author__ = "Ahmed Hermas"
__copyright__ = "Copyright 2021, © Digified "
__license__ = "MIT"
__version__ = "0.1.0"
__email__ = "a7medhermas@gmail.com"

from cv2 import cv2
import numpy as np


def _get_new_box(src_w, src_h, bbox, scale):
    """ change the new face boundries into based on the zoomout scale

        Args:
            src_w (int): Original Image width
            src_h (int): Original Image height
            bbox (tuple): face boundries in (xywh) format 
            scale (float): the zoomout scale

        Returns:
            [tuple]: the new face boundries
        """
    x = bbox[0]
    y = bbox[1]
    box_w = bbox[2]
    box_h = bbox[3]

    scale = min((src_h-1)/box_h, min((src_w-1)/box_w, scale))

    new_width = box_w * scale
    new_height = box_h * scale
    center_x, center_y = box_w/2+x, box_h/2+y

    left_top_x = center_x-new_width/2
    left_top_y = center_y-new_height/2
    right_bottom_x = center_x+new_width/2
    right_bottom_y = center_y+new_height/2

    if left_top_x < 0:
        right_bottom_x -= left_top_x
        left_top_x = 0

    if left_top_y < 0:
        right_bottom_y -= left_top_y
        left_top_y = 0

    if right_bottom_x > src_w-1:
        left_top_x -= right_bottom_x-src_w+1
        right_bottom_x = src_w-1

    if right_bottom_y > src_h-1:
        left_top_y -= right_bottom_y-src_h+1
        right_bottom_y = src_h-1

    return int(left_top_x), int(left_top_y),\
            int(right_bottom_x), int(right_bottom_y)

def crop(org_img, bbox, zoom_out, out_w=None, out_h=None):
    """ face cropping function,takes an image and return face cropped image

        Args:
            org_img (np array): Original Image
            bbox (tuple): face boundries in (xywh) format
            zoom_out (float): How zoomed out the face, 
            out_w (int): the output image width
            out_h (int): the output image height

        Returns:
            [np array]: the output face image
        """
    #get the shape of the image
    src_h, src_w, _ = np.shape(org_img)
    # get the new face boundries
    left_top_x, left_top_y, \
        right_bottom_x, right_bottom_y = _get_new_box(src_w, src_h, bbox, scale)
    #crop the image based on thge new boundries
    dst_img = org_img[left_top_y: right_bottom_y+1,
                    left_top_x: right_bottom_x+1]
    # check if you need to change the output image shape
    if out_w and out_h :
        
        dst_img = cv2.resize(img, (out_w, out_h))
    
    return dst_img


#!/usr/bin/env python3

"""detectors.py: contains face detectors modules."""

__author__ = "Ahmed Hermas"
__copyright__ = "Copyright 2021, © Digified "
__license__ = "MIT"
__version__ = "0.1.0"
__email__ = "a7medhermas@gmail.com"

from cv2 import cv2
import numpy as np
import math
import os
import logging
from .settings import MODELS_CONFIG

class Detector ():
    def __init__(self,model_name):
        
        self.model_name = model_name
        #calling the laoder function of the spicified model name
        self.model = getattr(self,MODELS_CONFIG[self.model_name]['loader'])()
    
    #Models Loaders 
    def _hog_cnn_loader(self):
        """ hog and cnn detector model
        Returns:
            None
        """
        global face_locations
        from face_recognition import face_locations 
        return None 
    
    def _cascade_loader(self):
        """ Cascade detector model

        Returns:
            [class]: detector class object
        """
        cascadePath = os.path.join(os.path.dirname(__file__),
                                   'assets/haarcascade_frontalface_default.xml')
        detector = cv2.CascadeClassifier(cascadePath)
        
        return detector

    def _retina_loader(self):
        """ retina detector model

        Returns:
            [class]: detector class object
        """
        #loading model paths
        caffemodel =os.path.join(os.path.dirname(__file__),
                                 "assets/Widerface-RetinaFace.caffemodel") 
        deploy = os.path.join(os.path.dirname(__file__),
                              "assets/deploy.prototxt")
        #loading the detector
        detector = cv2.dnn.readNetFromCaffe(deploy, caffemodel)
        
        return detector

    #Models Detectors
    def _getface_hog_cnn(self,img,mode):
        """ the detection function for cnn and hog models

        Args:
            img (np array): Input Image
            mode (int): 1 (xywh) , 2 (top,right,bottom,left)

        Returns:
            [tuple]: the output faces boundries
        """
        faces = face_locations(img,number_of_times_to_upsample=1,model=self.model_name)
        if len(faces)==0:
            return None
        if mode == 1:
            out = faces[0]
        elif mode ==2 :
            top,right,bottom,left = faces[0]
            x,y,w,h = int(left), int(top), int(right-left+1), int(bottom-top+1)
            out = [x,y,w,h]
        return out
    
    def _getface_cascade(self,img,mode):
        """the detection function for the cascade model

        Args:
            img (np array): Input Image
            mode (int): 1 (xywh) , 2(top,right,bottom,left)

        Returns:
            [tuple]: the output faces boundries
        """
        gray = cv2.cvtColor(img,cv2.COLOR_BGR2GRAY)
        faces = self.model.detectMultiScale(gray, 1.3, 5)
        if len(faces)==0:
            return None
        x,y,w,h = faces[0]
        top, right, bottom, left = self._change_shape(x,y,w,h)

        if mode == 1 :
            out = [x,y,w,h]
        elif mode == 2 :
            out = [top, right, bottom, left]
        return out

    def _getface_retina(self,img,mode):
        """ This is the retina face detection model implementation

        Args:
            img (np array): Input Image
            mode (int): 1 (xywh) , 2(top,right,bottom,left)

        Returns:
            [tuple]: the output faces boundries
        """
        height, width = img.shape[0], img.shape[1]
        aspect_ratio = width / height
        ####
        if img.shape[1] * img.shape[0] >= 192 * 192:
            img = cv2.resize(img,
                                (int(192 * math.sqrt(aspect_ratio)),
                                int(192 / math.sqrt(aspect_ratio))), interpolation=cv2.INTER_LINEAR)
        ####
        blob = cv2.dnn.blobFromImage(img, 1, mean=(104, 117, 123))
        self.model.setInput(blob, 'data')
        out = self.model.forward('detection_out').squeeze()
        max_conf_index = np.argmax(out[:, 2])
        if max_conf_index ==0.0:
            pass
        left, top, right, bottom = out[max_conf_index, 3]*width, out[max_conf_index, 4]*height, \
                                    out[max_conf_index, 5]*width, out[max_conf_index, 6]*height
        
        x,y,w,h = int(left), int(top), int(right-left+1), int(bottom-top+1)
        
        if mode == 1:
            out = [x,y,w,h]
        elif mode == 2:
            out = [top,right,bottom,left]
        return out
        
    def detect(self,image,mode):
        """ 
        detects faces in images and return the boundries
        Args:
            image (numpy array): the input image
            model (str): hog,cnn,cascade,retina
            mode (int): 1 (x,y,w,h) , 2 (top,right,bottom,left)

        Raises:
            ValueError: [4000]

        Returns:
            [tuple]: the output faces boundries
        """
        try:
            out =getattr(self,MODELS_CONFIG[self.model_name]['detector'])(image,mode)

        except UnboundLocalError as e :
            logging.exception("You Have entered a wrong mode number ") 
            
        return out
        
    def _change_shape(self,x,y,w,h):
        """ change output shape from xywh to top right bot left

        Args:
            x (int): [x coordinate of the top left corder]
            y (int): [y coordinate of the top left corder]
            w (int): [width of the Image]
            h (int): [height of the Image]

        Returns:
            [tuple]: new coordinates
        """
        top = y 
        left = x
        right = x + w
        bottom = y + h
        return top,right,bottom,left
